/*
 * $Id: sha1_crypt.c,v 1.2 2002/03/07 13:27:55 baggins Exp $
 */
#include "config.h"

#include <string.h>
#include "sha1.h"

static unsigned char itoa64[] =		/* 0 ... 63 => ascii - 64 */
	"./0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz";

static void
to64(char *s, unsigned long v, int n)
{
	while (--n >= 0) {
		*s++ = itoa64[v&0x3f];
		v >>= 6;
	}
}

/*
 * UNIX password
 *
 * Use SHA1 for what it is best at...
 */

char * crypt_sha1(const char *pw, const char *salt)
{
	const char *magic = "$3$";
	/* This string is magic for this algorithm.  Having
	 * it this way, we can get get better later on */
	char *passwd, *p;
	const char *sp,*ep;
	unsigned char	final[20];
	int sl,pl,i,j;
	SHA1_CTX	ctx,ctx1;
	unsigned long l;

	/* Refine the Salt first */
	sp = salt;

	/* TODO: now that we're using malloc'ed memory, get rid of the
	   strange constant buffer size. */
	passwd = malloc(120);

	/* If it starts with the magic string, then skip that */
	if(!strncmp(sp,magic,strlen(magic)))
		sp += strlen(magic);

	/* It stops at the first '$', max 8 chars */
	for(ep=sp;*ep && *ep != '$' && ep < (sp+8);ep++)
		continue;

	/* get the length of the true salt */
	sl = ep - sp;

	SHA1Init(&ctx);

	/* The password first, since that is what is most unknown */
	SHA1Update(&ctx,(unsigned const char *)pw,strlen(pw));

	/* Then our magic string */
	SHA1Update(&ctx,(unsigned const char *)magic,strlen(magic));

	/* Then the raw salt */
	SHA1Update(&ctx,(unsigned const char *)sp,sl);

	/* Then just as many characters of the SHA1(pw,salt,pw) */
	SHA1Init(&ctx1);
	SHA1Update(&ctx1,(unsigned const char *)pw,strlen(pw));
	SHA1Update(&ctx1,(unsigned const char *)sp,sl);
	SHA1Update(&ctx1,(unsigned const char *)pw,strlen(pw));
	SHA1Final(final,&ctx1);
	for(pl = strlen(pw); pl > 0; pl -= 20)
		SHA1Update(&ctx,(unsigned const char *)final,pl>20 ? 20 : pl);

	/* Don't leave anything around in vm they could use. */
	memset(final,0,sizeof final);

	/* Then something really weird... */
	for (j=0,i = strlen(pw); i ; i >>= 1)
		if(i&1)
		    SHA1Update(&ctx, (unsigned const char *)final+j, 1);
		else
		    SHA1Update(&ctx, (unsigned const char *)pw+j, 1);

	/* Now make the output string */
	strcpy(passwd,magic);
	strncat(passwd,sp,sl);
	strcat(passwd,"$");

	SHA1Final(final,&ctx);

	/*
	 * and now, just to make sure things don't run too fast
	 * On a 60 Mhz Pentium this takes 34 msec, so you would
	 * need 30 seconds to build a 1000 entry dictionary...
	 */
	for(i=0;i<1000;i++) {
		SHA1Init(&ctx1);
		if(i & 1)
			SHA1Update(&ctx1,(unsigned const char *)pw,strlen(pw));
		else
			SHA1Update(&ctx1,(unsigned const char *)final,20);

		if(i % 3)
			SHA1Update(&ctx1,(unsigned const char *)sp,sl);

		if(i % 7)
			SHA1Update(&ctx1,(unsigned const char *)pw,strlen(pw));

		if(i & 1)
			SHA1Update(&ctx1,(unsigned const char *)final,20);
		else
			SHA1Update(&ctx1,(unsigned const char *)pw,strlen(pw));
		SHA1Final(final,&ctx1);
	}

	p = passwd + strlen(passwd);

	l = (final[ 0]<<16) | (final[ 7]<<8) | final[14]; to64(p,l,4); p += 4;
	l = (final[ 1]<<16) | (final[ 8]<<8) | final[15]; to64(p,l,4); p += 4;
	l = (final[ 2]<<16) | (final[ 9]<<8) | final[16]; to64(p,l,4); p += 4;
	l = (final[ 3]<<16) | (final[10]<<8) | final[17]; to64(p,l,4); p += 4;
	l = (final[ 4]<<16) | (final[11]<<8) | final[18]; to64(p,l,4); p += 4;
	l = (final[ 5]<<16) | (final[12]<<8) | final[19]; to64(p,l,4); p += 4;
	l =                   (final[13]<<8) | final[ 6]; to64(p,l,3); p += 3;
	*p = '\0';

	/* Don't leave anything around in vm they could use. */
	memset(final,0,sizeof final);

	return passwd;
}

